from typing import Optional
from pydantic import BaseModel
from abc import ABC, abstractmethod
from collections.abc import Generator, AsyncGenerator


class ChatTokenUsage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class ChatResponse(BaseModel):
    content: str
    token_usage: Optional[ChatTokenUsage]


class StreamedChatResponse(Generator[ChatResponse, None, None], ABC):
    @abstractmethod
    def generate_chat_response(self) -> ChatResponse:
        """Generate the next chat response.
        Raise `StopIteration` if there are no more responses.

        Returns
        -------
        ChatResponse
            The next chat response.
        """

        pass

    def send(self, value) -> ChatResponse:
        """Send a value into the generator.
        Return next yielded value or raise StopIteration.
        """

        return self.generate_chat_response()

    def throw(self, typ, val=None, tb=None):
        """Raise an exception in the generator.
        Return next yielded value or raise StopIteration.

        Notes
        -----
        The implementation of this method is taken from the CPython source code:
        https://github.com/python/cpython/blob/d5d3249e8a37936d32266fa06ac20017307a1f70/Lib/_collections_abc.py#L309
        """

        if val is None:
            if tb is None:
                raise typ
            val = typ()
        if tb is not None:
            val = val.with_traceback(tb)
        raise val


class AsyncStreamedChatResponse(AsyncGenerator[ChatResponse, None, None], ABC):
    @abstractmethod
    async def generate_chat_response_async(self) -> ChatResponse:
        """Generate the next chat response asynchronously.
        Raise `StopAsyncIteration` if there are no more responses.

        Returns
        -------
        ChatResponse
            The next chat response.
        """

        pass

    async def asend(self, value) -> ChatResponse:
        """Send a value into the asynchronous generator.
        Return next yielded value or raise StopAsyncIteration.
        """

        return await self.generate_chat_response_async()

    async def athrow(self, typ, val=None, tb=None):
        """Raise an exception in the asynchronous generator.
        Return next yielded value or raise StopAsyncIteration.

        Notes
        -----
        The implementation of this method is taken from the CPython source code:
        https://github.com/python/cpython/blob/d5d3249e8a37936d32266fa06ac20017307a1f70/Lib/_collections_abc.py#L309
        """

        if val is None:
            if tb is None:
                raise typ
            val = typ()
        if tb is not None:
            val = val.with_traceback(tb)
        raise val



from .chat import ChatMessage, UserChatMessage, AssistantChatMessage
from .openai import OpenAIChatModel
from .qianfan import QianfanChatModel

__all__ = [
    "ChatMessage",
    "UserChatMessage",
    "AssistantChatMessage",
    "OpenAIChatModel",
    "QianfanChatModel",
]



from .model import ChatModel
from .message import ChatMessage, UserChatMessage, AssistantChatMessage, ChatRole
from .response import (
    ChatResponse,
    ChatTokenUsage,
    StreamedChatResponse,
    AsyncStreamedChatResponse,
)

__all__ = [
    "ChatModel",
    "ChatMessage",
    "UserChatMessage",
    "AssistantChatMessage",
    "ChatRole",
    "ChatResponse",
    "ChatTokenUsage",
    "StreamedChatResponse",
    "AsyncStreamedChatResponse",
]



import qianfan

from .auth import QianfanAuth
from ..chat import (
    ChatModel,
    ChatMessage,
    ChatResponse,
    ChatTokenUsage,
    StreamedChatResponse,
    UserChatMessage,
)
from .response import QianfanStreamedChatResponse, AsyncQianfanStreamedChatResponse


class QianfanChatModel(ChatModel):
    auth: QianfanAuth = QianfanAuth()

    def get_complete_response(
        self,
        messages: str | ChatMessage | list[ChatMessage],
    ) -> ChatResponse:
        # Convert to a list of messages
        messages = self._convert_to_chat_message_list(messages)

        # Call the API
        response = qianfan.ChatCompletion(
            model=self.name,
            ak=self.auth.access_key,
            sk=self.auth.secret_key,
        ).do(
            messages=list(
                map(
                    lambda message: dict(
                        role=message.role.value,
                        content=message.content,
                    ),
                    messages,
                )
            ),
            temperature=self.temperature,
        )

        # Convert a ChatResponse
        response = ChatResponse(
            content=response["result"],
            token_usage=ChatTokenUsage(
                prompt_tokens=response["usage"]["prompt_tokens"],
                completion_tokens=response["usage"]["completion_tokens"],
                total_tokens=response["usage"]["total_tokens"],
            ),
        )

        return response

    def get_streamed_response(
        self,
        messages: str | ChatMessage | list[ChatMessage],
    ) -> StreamedChatResponse:
        # Convert to a list of messages
        messages = self._convert_to_chat_message_list(messages)

        # Call the API
        iterator = qianfan.ChatCompletion(
            model=self.name,
            ak=self.auth.access_key,
            sk=self.auth.secret_key,
        ).do(
            messages=list(
                map(
                    lambda message: dict(
                        role=message.role.value,
                        content=message.content,
                    ),
                    messages,
                )
            ),
            temperature=self.temperature,
            # Enable streaming
            stream=True,
        )

        # Convert to streamed response
        streamed_response = QianfanStreamedChatResponse(iterator)

        return streamed_response

    async def get_complete_response_async(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> ChatResponse:
        # Convert to a list of messages
        messages = self._convert_to_chat_message_list(messages)

        # Call the API
        response = await qianfan.ChatCompletion(
            model=self.name,
            ak=self.auth.access_key,
            sk=self.auth.secret_key,
        ).ado(
            messages=list(
                map(
                    lambda message: dict(
                        role=message.role.value,
                        content=message.content,
                    ),
                    messages,
                )
            ),
            temperature=self.temperature,
        )

        # Convert a ChatResponse
        response = ChatResponse(
            content=response["result"],
            token_usage=ChatTokenUsage(
                prompt_tokens=response["usage"]["prompt_tokens"],
                completion_tokens=response["usage"]["completion_tokens"],
                total_tokens=response["usage"]["total_tokens"],
            ),
        )

        return response

    async def get_streamed_response_async(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> AsyncQianfanStreamedChatResponse:
        # Convert to a list of messages
        messages = self._convert_to_chat_message_list(messages)

        # Call the API
        stream = await qianfan.ChatCompletion(
            model=self.name,
            ak=self.auth.access_key,
            sk=self.auth.secret_key,
        ).ado(
            messages=list(
                map(
                    lambda message: dict(
                        role=message.role.value,
                        content=message.content,
                    ),
                    messages,
                )
            ),
            temperature=self.temperature,
            # Enable streaming
            stream=True,
        )

        # Convert to streamed response
        streamed_response = AsyncQianfanStreamedChatResponse(stream)

        return streamed_response



import openai

from ..chat import ChatResponse, StreamedChatResponse, AsyncStreamedChatResponse


class OpenAIStreamedChatResponse(StreamedChatResponse):
    def __init__(self, stream: openai.Stream) -> None:
        # Store OpenAI's stream
        self._stream = stream

    def generate_chat_response(self) -> ChatResponse:
        # Get the next chunk
        chunk = next(self._stream)

        # Extract the content
        content = chunk.choices[0].delta.content

        if content is not None:
            return ChatResponse(
                content=content,
                token_usage=None,
            )
        else:
            raise StopIteration


class AsyncOpenAIStreamedChatResponse(AsyncStreamedChatResponse):
    def __init__(self, stream: openai.AsyncStream) -> None:
        # Store OpenAI's async stream
        self._stream = stream

    async def generate_chat_response_async(self) -> ChatResponse:
        # Get the next chunk
        chunk = await anext(self._stream)

        # Extract the content
        content = chunk.choices[0].delta.content

        if content is not None:
            return ChatResponse(
                content=content,
                token_usage=None,
            )
        else:
            raise StopAsyncIteration



from abc import ABC, abstractmethod
from pydantic import BaseModel
from ..auth import LLMAuth
from .message import ChatMessage, UserChatMessage
from .response import ChatResponse, StreamedChatResponse


class ChatModel(BaseModel, ABC):
    name: str
    temperature: float = 0.9
    auth: LLMAuth

    @abstractmethod
    def get_complete_response(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> ChatResponse:
        """Get a complete response from the LLM chat model.

        Parameters
        ----------
        messages : str | ChatMessage | list[ChatMessage]
            One of the following:
            - User message content.
            - A single user message.
            - A list of messages of the user and the assistant.

        Returns
        -------
        ChatResponse
            A response from the LLM chat model.
        """

        pass

    @abstractmethod
    def get_streamed_response(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> StreamedChatResponse:
        """Get a streamed response from the LLM chat model.

        Parameters
        ----------
        messages : str | ChatMessage | list[ChatMessage]
            One of the following:
            - User message content.
            - A single user message.
            - A list of messages of the user and the assistant.

        Returns
        -------
        StreamedChatResponse
            A streamed response from the LLM chat model.
        """
        pass

    @abstractmethod
    async def get_complete_response_async(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> ChatResponse:
        """Get a complete response from the LLM chat model asynchronously.

        Parameters
        ----------
        messages: str | UserChatMessage | list[ChatMessage]
            One of the following:
            - User message content.
            - A single user message.
            - A list of messages of the user and the assistant.

        Returns
        -------
        ChatResponse
            A response from the LLM chat model.
        """

        pass

    @abstractmethod
    async def get_streamed_response_async(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> StreamedChatResponse:
        """Get a streamed response from the LLM chat model asynchronously.

        Parameters
        ----------
        messages: str | UserChatMessage | list[ChatMessage]
            One of the following:
            - User message content.
            - A single user message.
            - A list of messages of the user and the assistant.

        Returns
        -------
        StreamedChatResponse
            A streamed response from the LLM chat model.
        """

        pass

    @staticmethod
    def _convert_to_chat_message_list(
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> list[ChatMessage]:
        """Convert a single user message or a string to a list of messages
        if the input `messages` is actually a single message.

        Parameters
        ----------
        messages : str | UserChatMessage | list[ChatMessage]
            One of the following:
            - User message content.
            - A single user message.
            - A list of messages of the user and the assistant.

        Returns
        -------
        list[ChatMessage]
            A list of messages.
        """

        # Convert user message content to a list of messages
        if isinstance(messages, str):
            messages = [UserChatMessage(content=messages)]

        # Convert a single user message to a list
        if isinstance(messages, UserChatMessage):
            messages = [messages]

        return messages



from typing import Coroutine, Iterator
import qianfan

from ..chat import (
    ChatResponse,
    ChatTokenUsage,
    StreamedChatResponse,
    AsyncStreamedChatResponse,
)


class QianfanStreamedChatResponse(StreamedChatResponse):
    def __init__(self, iterator: Iterator[qianfan.QfResponse]) -> None:
        # Qianfan's streamed response as an iterator
        self._iterator = iterator

    def generate_chat_response(self) -> ChatResponse:
        # Get the next chunk
        chunk = next(self._iterator)

        # Create a ChatResponse
        response = ChatResponse(
            content=chunk["result"],
            token_usage=ChatTokenUsage(
                prompt_tokens=chunk["usage"]["prompt_tokens"],
                completion_tokens=chunk["usage"]["completion_tokens"],
                total_tokens=chunk["usage"]["total_tokens"],
            ),
        )

        return response


class AsyncQianfanStreamedChatResponse(AsyncStreamedChatResponse):
    def __init__(self, stream: Coroutine) -> None:
        # Qianfan's streamed response as an iterator
        self._stream = stream

    async def generate_chat_response_async(self) -> ChatResponse:
        # Get the next chunk
        chunk = await anext(self._stream)

        # Create a ChatResponse
        response = ChatResponse(
            content=chunk["result"],
            token_usage=ChatTokenUsage(
                prompt_tokens=chunk["usage"]["prompt_tokens"],
                completion_tokens=chunk["usage"]["completion_tokens"],
                total_tokens=chunk["usage"]["total_tokens"],
            ),
        )

        return response



from enum import StrEnum
from pydantic import BaseModel, ConfigDict


class ChatRole(StrEnum):
    USER = "user"
    ASSISTANT = "assistant"


class ChatMessage(BaseModel):
    role: ChatRole
    content: str

    model_config = ConfigDict(
        frozen=True,
    )


class UserChatMessage(ChatMessage):
    role: ChatRole = ChatRole.USER


class AssistantChatMessage(ChatMessage):
    role: ChatRole = ChatRole.ASSISTANT



from openai import OpenAI, AsyncOpenAI

from .auth import OpenAIAuth
from ..chat import (
    ChatModel,
    ChatMessage,
    UserChatMessage,
    ChatResponse,
    ChatTokenUsage,
)
from .response import OpenAIStreamedChatResponse, AsyncOpenAIStreamedChatResponse


class OpenAIChatModel(ChatModel):
    auth: OpenAIAuth = OpenAIAuth()

    def get_complete_response(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> ChatResponse:
        # Convert a single message to a list
        messages = self._convert_to_chat_message_list(messages)

        # Call the API
        response = OpenAI(api_key=self.auth.api_key).chat.completions.create(
            model=self.name,
            messages=list(
                map(
                    lambda message: dict(
                        role=message.role.value,
                        content=message.content,
                    ),
                    messages,
                )
            ),
            temperature=self.temperature,
        )

        # Convert a ChatResponse
        response = ChatResponse(
            content=response.choices[0].message.content,
            token_usage=ChatTokenUsage(
                prompt_tokens=response.usage.prompt_tokens,
                completion_tokens=response.usage.completion_tokens,
                total_tokens=response.usage.total_tokens,
            ),
        )

        return response

    def get_streamed_response(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> OpenAIStreamedChatResponse:
        # Convert a single message to a list
        messages = self._convert_to_chat_message_list(messages)

        # Call the API and get the stream
        stream = OpenAI(api_key=self.auth.api_key).chat.completions.create(
            model=self.name,
            messages=list(
                map(
                    lambda message: dict(
                        role=message.role.value,
                        content=message.content,
                    ),
                    messages,
                )
            ),
            temperature=self.temperature,
            # Enable streaming
            stream=True,
        )

        # Create a streamed response
        streamed_response = OpenAIStreamedChatResponse(stream)

        return streamed_response

    async def get_complete_response_async(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> ChatResponse:
        # Convert a single message to a list
        messages = self._convert_to_chat_message_list(messages)

        # Call the API
        response = await AsyncOpenAI(api_key=self.auth.api_key).chat.completions.create(
            model=self.name,
            messages=list(
                map(
                    lambda message: dict(
                        role=message.role.value,
                        content=message.content,
                    ),
                    messages,
                )
            ),
            temperature=self.temperature,
        )

        # Convert a ChatResponse
        response = ChatResponse(
            content=response.choices[0].message.content,
            token_usage=ChatTokenUsage(
                prompt_tokens=response.usage.prompt_tokens,
                completion_tokens=response.usage.completion_tokens,
                total_tokens=response.usage.total_tokens,
            ),
        )

        return response

    async def get_streamed_response_async(
        self,
        messages: str | UserChatMessage | list[ChatMessage],
    ) -> AsyncOpenAIStreamedChatResponse:
        # Convert a single message to a list
        messages = self._convert_to_chat_message_list(messages)

        # Call the API
        stream = await AsyncOpenAI(api_key=self.auth.api_key).chat.completions.create(
            model=self.name,
            messages=list(
                map(
                    lambda message: dict(
                        role=message.role.value,
                        content=message.content,
                    ),
                    messages,
                )
            ),
            temperature=self.temperature,
            # Enable streaming
            stream=True,
        )

        # Create a streamed response
        streamed_response = AsyncOpenAIStreamedChatResponse(stream)

        return streamed_response



from typing import Self
from abc import ABC, abstractclassmethod
from pydantic_settings import BaseSettings


class LLMAuth(BaseSettings, ABC):
    pass
    # @abstractclassmethod
    # def from_env(cls) -> Self:
    #     pass
